package jossihawk;

import jossihawk.ep.AIMProblemResult;
import jossihawk.ep.ChosenIndividual;
import jossihawk.ep.MOEARunStatistics;
import jossihawk.ep.SingleAIMProblemRunner;
import jossihawk.simulator.AIMSimulator;
import jossihawk.simulator.AIMStats;
import jossihawk.simulator.SimulationProperties;
import jossihawk.simulator.SimulationState;
import jossihawk.simulator.sim.MOEASimulationState;
import jossihawk.simulator.sim.playback.ScenarioStorer;
import jossihawk.simulator.sim.playback.SimulationPlayback;
import jossihawk.simulator.sim.playback.SimulationRecordingKeeper;
import jossihawk.utils.ChosenIndividualChartDrawer;
import jossihawk.utils.HistogramChartDrawer;
import jossihawk.utils.ScalingDataPoint;
import jossihawk.utils.ShutdownHook;
import org.apache.log4j.*;

import javax.xml.bind.JAXBException;
import java.io.*;
import java.util.ArrayList;
import java.util.List;
import java.util.Scanner;

public class ExperimentRunner {

    final static Logger logger = Logger.getLogger(ExperimentRunner.class);
    static  String logfile;
    static {
        Scanner scan = new Scanner(System.in);
        System.out.println("Hey! Who are you? (please type the same name here each time, also no funny chars):");
        logfile = "logs/" + scan.nextLine() + ".log";
        System.out.println("Thanks for running our thesis code :)");

        ConsoleAppender console = new ConsoleAppender(); //create appender
        //configure the appender
        String PATTERN = "%d [%p|%c|%C{1}] %m%n";
        console.setLayout(new PatternLayout(PATTERN));
        console.setThreshold(Level.DEBUG);
        console.activateOptions();
        //put appender to any Logger (here is root)
        Logger.getRootLogger().addAppender(console);

        FileAppender fa = new FileAppender();
        fa.setName("FileLogger");
        fa.setFile(logfile);
        fa.setLayout(new PatternLayout("%d %-5p [%c{1}] %m%n"));
        fa.setThreshold(Level.DEBUG);
        fa.setAppend(true);
        fa.activateOptions();

        //put appender to any Logger (here is root)
        Logger.getRootLogger().addAppender(fa);

        logger.debug("NEW RUN!");
        logger.debug("NEW RUN!");
        Runtime.getRuntime().addShutdownHook(new ShutdownHook());
    }

    private AIMSimulator experimentSimulator;
    private SimulationProperties initialSimulationProperties;
    private ChosenIndividualChartDrawer chosenIndividualChartDrawer;
    private HistogramChartDrawer metHistogramDrawer;
    private SimulationPlayback playback;
    private SimulationRecordingKeeper keeper;
    private List<ScalingDataPoint> scalingDataPoints;
    private List<MOEARunStatistics> moeaStatisticsList;


    /**
     * This constructor is not guaranteed to work, use other constructor. (Used for testing purposes)
     * @param propertiesFilepath
     * @throws FileNotFoundException
     * @throws JAXBException
     */
    public ExperimentRunner(String propertiesFilepath) throws FileNotFoundException, JAXBException {
        initialSimulationProperties = new SimulationProperties(propertiesFilepath);

        //Determine if we should load from scenario or not
        if (!initialSimulationProperties.isStartScenario()) {
            experimentSimulator = AIMSimulator.getInstance(new SimulationState(null, initialSimulationProperties));
        } else {
            MOEASimulationState startingState = ScenarioStorer.getStateFromFile(new File(ScenarioStorer.replayFolder + "state.xml"));
            experimentSimulator = AIMSimulator.getInstance(new SimulationState(startingState.getState().getVehicleStates(), initialSimulationProperties));
        }
        if (initialSimulationProperties.getGraphChosenIndividual())
            chosenIndividualChartDrawer = new ChosenIndividualChartDrawer(initialSimulationProperties);
        if (initialSimulationProperties.isShowMETHistogram())
            metHistogramDrawer = new HistogramChartDrawer("MET Histogram");
        experimentSimulator.setAsMainSimulator(initialSimulationProperties.isHeadless());
        experimentSimulator.setSpawnVehicles(initialSimulationProperties.isSpawnVehiclesInMain());

        if (initialSimulationProperties.isRecordForPlayback()) {
            playback = new SimulationPlayback();
            try {
                keeper = new SimulationRecordingKeeper(initialSimulationProperties.isImportantRun(), initialSimulationProperties.getTitle());
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }

    public ExperimentRunner(File scenarioDirectory) throws FileNotFoundException, JAXBException {
        initialSimulationProperties = new SimulationProperties(scenarioDirectory.getAbsolutePath() + File.separator + scenarioDirectory.getName() + ".properties");
        //Determine if we should load from scenario or not
        if (!initialSimulationProperties.isStartScenario()) {
            experimentSimulator = AIMSimulator.getInstance(new SimulationState(null, initialSimulationProperties));
        } else {
            MOEASimulationState startingState = ScenarioStorer.getStateFromFile(new File(scenarioDirectory, "state.xml"));
            experimentSimulator = AIMSimulator.getInstance(new SimulationState(startingState.getState().getVehicleStates(), initialSimulationProperties));
        }
        if (initialSimulationProperties.getGraphChosenIndividual())
            chosenIndividualChartDrawer = new ChosenIndividualChartDrawer(initialSimulationProperties);
        if (initialSimulationProperties.isShowMETHistogram())
            metHistogramDrawer = new HistogramChartDrawer("MET Histogram");
        experimentSimulator.setAsMainSimulator(initialSimulationProperties.isHeadless());
        experimentSimulator.setSpawnVehicles(initialSimulationProperties.isSpawnVehiclesInMain());

        if (initialSimulationProperties.isRecordForPlayback()) {
            playback = new SimulationPlayback();
            try {
                keeper = new SimulationRecordingKeeper(initialSimulationProperties.isImportantRun(), initialSimulationProperties.getTitle());
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }

    public AIMSimulator run() {
        scalingDataPoints = new ArrayList<>();
        moeaStatisticsList = new ArrayList<>();

        int step = 0;
        SimulationState simulationState = experimentSimulator.getState();
        if(initialSimulationProperties.getAlgorithmTimeStepSize() > initialSimulationProperties.getNumberOfSteps())
            throw new RuntimeException("Step size bigger than simulation time.");

        while (step < initialSimulationProperties.getNumberOfSteps() / initialSimulationProperties.getAlgorithmTimeStepSize()) {
            ChosenIndividual chosenIndividual = null;
            if (simulationState.getSimulationProperties().getNumberOfCars() > 0) {
                SingleAIMProblemRunner singleAIMProblemRunner = new SingleAIMProblemRunner(simulationState);
                AIMProblemResult result = singleAIMProblemRunner.run();
                if(result.getScalingDataPoint() != null)
                    scalingDataPoints.add(result.getScalingDataPoint());
                chosenIndividual = result.getChosenIndividual();

                if(initialSimulationProperties.isRecordMoeaRunStatistics()) {
                    moeaStatisticsList.add(singleAIMProblemRunner.getTerminationCondition().getMoeaRunStatistics());
                }
            } else chosenIndividual = new ChosenIndividual(null, 0);

            double[] speeds = chosenIndividual.getSpeeds();

            if (initialSimulationProperties.isRecordForPlayback())
                recordState(simulationState, speeds, experimentSimulator.getSimulationTime());
            /*AIMSimulator tempSimController = AIMSimulator.getInstance(experimentSimulator.getState());
            tempSimController.setAsMainSimulator();
            tempSimController.evaluate(chosenIndividual.getSpeeds());*/
            experimentSimulator.step(speeds);
            if(experimentSimulator.getNumberOfActiveCars() == 0 && !initialSimulationProperties.isSpawnVehiclesInMain())
                break;

            if (initialSimulationProperties.getGraphChosenIndividual()) {
                chosenIndividualChartDrawer.addChosenIndividual(step, chosenIndividual);
            }
            simulationState = experimentSimulator.getState();
            step++;

            //Get global stats
            AIMStats globalStats = experimentSimulator.getTotalStats();
            
            if (initialSimulationProperties.isShowMETHistogram())

                metHistogramDrawer.addData(globalStats.getVehicleStatisticsMap().values());
            if (initialSimulationProperties.isScenarioCreation()) { //goal is scenario creation
                if (experimentSimulator.getNumberOfActiveCars() > initialSimulationProperties.getCarsInScenraio()) { //we have enough cars to stop
                    storeScenario(speeds);
                    logger.debug("Scenario complete");
                    break;
                }
            }

        }
        return experimentSimulator;
    }

    private void storeScenario(double[] speeds) {
        try {
            ScenarioStorer storer = new ScenarioStorer(initialSimulationProperties.getCarsInScenraio(), initialSimulationProperties.getSpawnRateSouth(), initialSimulationProperties.getSpawnRateNorth(), initialSimulationProperties.getSpawnRateEast(), initialSimulationProperties.getSpawnRateWest());
            SimulationPlayback playback = new SimulationPlayback();
            MOEASimulationState moeaState = new MOEASimulationState(experimentSimulator.getState(), speeds, experimentSimulator.getSimulationTime());
            playback.addState(moeaState);
            storer.recordSimulationPlayback(playback);
        } catch (IOException e) {
            e.printStackTrace();
            logger.debug("Failed writing scenario to file");
        } catch (JAXBException e) {
            e.printStackTrace();
            logger.debug("Failed writing scenario to file");
        }
    }

    private void recordState(SimulationState state, double[] speedVector, double simulationTime) {
        MOEASimulationState moeaState = new MOEASimulationState(state, speedVector, simulationTime);
        playback.addState(moeaState);
        try {
            keeper.recordSimulationPlayback(playback);
        } catch (JAXBException e) {
            e.printStackTrace();
        }
    }

    public void writeCurrentPlaybackToFile(File file) {
        try {
            keeper.recordSimulationPlayback(playback, file);
        } catch (JAXBException e) {
            e.printStackTrace();
        }
    }

    public static void run_test(String propertiesFile) throws FileNotFoundException, JAXBException {
        ExperimentRunner experimentRunner = new ExperimentRunner(propertiesFile);
        AIMSimulator simulator = experimentRunner.run();
    }

    public static void run_experiments(int experimentNumber) throws JAXBException, FileNotFoundException {
        // List all experiments and prompt user to select one
        //ExperimentRunner experimentRunner = new ExperimentRunner("test.properties");
        //AIMSimulator simulator = experimentRunner.run();
        String experimentsFolderPath = "experiments";
        File experimentsFolder = new File(experimentsFolderPath);
        String[] experimentStrings = experimentsFolder.list(new FilenameFilter() {
            @Override
            public boolean accept(File dir, String name) {
                return new File(dir, name).isDirectory();
            }
        });
        if(experimentStrings == null || experimentStrings.length == 0)
            logger.error("No experiments found in directory: " + experimentsFolder.getAbsolutePath());
        for (int i = 0; i < experimentStrings.length; i++) {
            System.out.println(i + ":   " + experimentStrings[i]);
        }

        int experimentIndex = experimentNumber;
        if (experimentNumber == -1) {
            Scanner scanner = new Scanner(System.in);
            System.out.print("Enter experiment number: ");
            experimentIndex = scanner.nextInt();
        }

        // Run scenarions from chosen experiment
        String experimentFolderName = experimentStrings[experimentIndex];
        File experimentFile = new File(experimentsFolderPath, experimentFolderName);
        logger.debug("Looking for scenarios in " + experimentFile.getAbsolutePath());
        File[] scenarios = experimentFile.listFiles(new FileFilter() {
            @Override
            public boolean accept(File pathname) {
                return pathname.isDirectory();
            }
        });
        logger.debug("Found " + scenarios.length + " scenarios\n");
        for (int i = 0; i < scenarios.length; i++) {
            File scenario = scenarios[i];
            File resultsFile = new File(scenario, "results");
            if (!resultsFile.exists()) {
                resultsFile.mkdir();
                ExperimentRunner experimentRunner = new ExperimentRunner(scenario);
                logger.debug("Running scenario: " + scenario.getAbsolutePath());
                AIMSimulator simulator = experimentRunner.run();
                experimentRunner.writeCurrentPlaybackToFile(new File(resultsFile, "playback.xml"));
                logger.debug("Finished scenario: " + scenario.getAbsolutePath());
                AIMStats globalStats = simulator.getTotalStats();
                globalStats.setScalingDataPoints(experimentRunner.getScalingDataPoints());
                globalStats.setMoeaRunStatistics(experimentRunner.getMoeaStatisticsList());
                File resultFile = new File(resultsFile, "globalStats.xml");
                logger.debug("Writing results to: " + resultFile.getAbsolutePath());
                logger.debug("Result: " + globalStats.toString());
                logger.debug("");
                globalStats.writeToFile(resultFile);
                //logger.debug(AIMStats.readFromFile(resultFile).toString());
            } else {
                logger.debug("Skipped scenario: " + scenario.getAbsolutePath() + "\n");
            }
        }
    }

    public static void main(String[] args) throws FileNotFoundException, JAXBException {
//        run_test("test.properties");
        if(args.length != 0)
            run_experiments(Integer.parseInt(args[0]));
        else
            run_experiments(-1);
    }

    public List<ScalingDataPoint> getScalingDataPoints() {
        return scalingDataPoints;
    }
    public List<MOEARunStatistics> getMoeaStatisticsList() {
        return moeaStatisticsList;
    }
}
