package jossihawk.ep;

import jossihawk.simulator.AIMSimulator;
import jossihawk.simulator.SimulationProperties;
import jossihawk.simulator.SimulationState;
import org.moeaframework.core.Solution;
import org.moeaframework.core.variable.EncodingUtils;
import org.moeaframework.core.variable.RealVariable;
import org.moeaframework.problem.AbstractProblem;

public class AIMProblem extends AbstractProblem {

    private final SimulationState simulationState;
    private final double bestTinyThroughput, worstStarvation;
    private SimulationProperties simulationProperties;
    private double maxKineticEnergy, maxDistanceTravelled;

    /**
     * This class specifies the Autonomous Intersection Management Problem.
     *
     * @param simulationState provides the variables relevant for this problem instance.
     */
    public AIMProblem(SimulationState simulationState, double maxKineticEnergy, double maxDistanceTravelled, double bestTinyThroughput, double worstStarvation) {
        super(simulationState.getSimulationProperties().getNumberOfCars(), simulationState.getSimulationProperties().getNumberOfObjectives(), 1);
        this.simulationProperties = simulationState.getSimulationProperties();
        this.simulationState = simulationState;
        this.maxKineticEnergy = maxKineticEnergy;
        this.maxDistanceTravelled = maxDistanceTravelled;
        this.bestTinyThroughput = bestTinyThroughput;
        this.worstStarvation = worstStarvation;
    }

    public AIMProblem(SimulationState simulationState, int nofVariables, int nofObjectives, int nofConstraints, double maxKineticEnergy, double maxDistanceTravelled, double bestTinyThroughput, double worstStarvation) {
        super(nofVariables, nofObjectives, nofConstraints);
        this.simulationState = simulationState;
        this.simulationProperties = simulationState.getSimulationProperties();
        this.maxKineticEnergy = maxKineticEnergy;
        this.maxDistanceTravelled = maxDistanceTravelled;
        this.bestTinyThroughput = bestTinyThroughput;
        this.worstStarvation = worstStarvation;
    }

    protected SimulationProperties getSimulationProperties() {
        return simulationProperties;
    }

    public void evaluate(Solution solution) {
        AIMSimulator simulator = AIMSimulator.getInstance(simulationState);
        double nofCars = (double) getNumberOfVariables();
        double[] speeds = EncodingUtils.getReal(solution);
        simulator.evaluate(speeds);

        if (simulationProperties.getConstrainCollisions()) {
            double collisions = simulator.getNumberOfCollisions();
            solution.setConstraint(0, collisions <= 0.0 ? 0.0 : collisions);
        }

        int objective_idx = 0;
        if (simulationProperties.getEvaluateCollisions()) {
            solution.setObjective(objective_idx, simulator.getNumberOfCollisions()/nofCars);
            objective_idx++;
        }

        if (simulationProperties.isEvaluateDistanceTravelled()) {
            solution.setObjective(objective_idx, -simulator.getTotalDistanceTraveledInLastTimeStep()/maxDistanceTravelled);
            objective_idx++;
        }

        if (simulationProperties.getEvaluateThroughput()) {
            solution.setObjective(objective_idx, -simulator.getThroughput() / nofCars);
            objective_idx++;
        }

        if (simulationProperties.isEvaluateEk()) {
            solution.setObjective(objective_idx, simulator.getKineticEnergyLostInLastTimeStep()/maxKineticEnergy);
            objective_idx++;
        }

        if (simulationProperties.isEvaluateStarvation()) {
            solution.setObjective(objective_idx, simulator.getStoppageTimeOfStoppedCars()/worstStarvation);
            objective_idx++;
        }

        if (simulationProperties.isEvaluateTinyThroughput()) {
            double ttDivisor = bestTinyThroughput == 0.0 ? 1 : bestTinyThroughput;
            solution.setObjective(objective_idx, -simulator.getCarsEvacuatedFromInnerIntersectionInLastTimeStep() / ttDivisor);
            objective_idx++;
        }
        if(simulationProperties.isEvaluateInvertedDistanceTravelled()) {
            solution.setObjective(objective_idx, simulator.getTotalDistanceTraveledInLastTimeStep()/maxDistanceTravelled);
            objective_idx++;
        }
    }

    public Solution newSolution() {
        Solution solution = new Solution(getNumberOfVariables(), getNumberOfObjectives(), getNumberOfConstraints());
        for (int i = 0; i < getNumberOfVariables(); i++) {
            RealVariable var = new RealVariable(0, simulationProperties.getSpeedLimit());
            solution.setVariable(i, var);
        }
        return solution;
    }

}
