package jossihawk.simulator.sim;

import aim4.config.Debug;
import aim4.driver.AutoDriver;
import aim4.im.RoadBasedIntersection;
import aim4.map.BasicMap;
import aim4.map.DataCollectionLine;
import aim4.map.SpawnPoint;
import aim4.map.lane.Lane;
import aim4.sim.Simulator;
import aim4.vehicle.*;
import jossihawk.simulator.sim.stats.SmallScaleStatisticsEventListener;

import java.awt.*;
import java.awt.geom.Area;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.*;
import java.util.List;
import java.util.stream.Collectors;

/**
 * Adaptation of AutoDriverOnlySimulator
 */
public class MOEASimulator implements Simulator {

    ThreadSafeVinRegistry registry = new ThreadSafeVinRegistry();
    List<SimulatorStateChangeListener> stateChangeListeners = new ArrayList<SimulatorStateChangeListener>(); //Listeners for GUI level access
    List<SmallScaleStatisticsEventListener> statisticsEventListeners = new ArrayList<SmallScaleStatisticsEventListener>();
    public final RoadBasedIntersection intersection;
    private Set<Integer> carsOutOfInnerIntersection = new HashSet<>();
    Map<Integer,Double> stoppageTimeOfAllCars = new HashMap<>();
    private double distanceTraveled = 0;
    public double timeLeft = 0.001;

    public MOEASimulator(BasicMap basicMap) {
        this.basicMap = basicMap;
        this.spawnVehicles = false;
        this.vinToVehicles = new HashMap<Integer, VehicleSimView>();

        List roads = basicMap.getRoads();
        intersection = new RoadBasedIntersection(roads);

        currentTime = 0.0;
        numOfCompletedVehicles = 0;

    }

    /**
     * The map
     */
    private BasicMap basicMap;
    /**
     * Spawn vehicles?
     */
    private boolean spawnVehicles;
    /**
     * All active vehicles, in form of a map from VINs to vehicle objects.
     */
    private Map<Integer, VehicleSimView> vinToVehicles;
    /**
     * The current time
     */
    private double currentTime;
    /**
     * The number of completed vehicles
     */
    private int numOfCompletedVehicles;

    public double getStoppageTimeOfStoppedCars() {
        return stoppageTimeOfAllCars
                .keySet()
                .stream()
                .filter(key -> getVinsOfStoppedCars(0.01).contains(key))
                .mapToDouble(key -> stoppageTimeOfAllCars.get(key))
                .sum();
    }


    public static class MOEASimStepResult implements CollidableSimStepResult {

        /**
         * The VIN of the completed vehicles in this time step
         */
        private Set<Integer> spawnedVINs;
        List<Integer> completedVINs;
        Set<Integer> crashedVINs;

        /**
         * Create a result of a simulation step
         *
         * @param completedVINs the VINs of completed vehicles.
         */
        public MOEASimStepResult(List<Integer> completedVINs, Set<Integer> crashedVINs, Set<Integer> spawnedVINs) {
            this.completedVINs = completedVINs;
            this.crashedVINs = crashedVINs;
            this.spawnedVINs = spawnedVINs;
        }

        /**
         * Get the list of VINs of completed vehicles.
         *
         * @return the list of VINs of completed vehicles.
         */
        public List<Integer> getCompletedVINs() {
            return completedVINs;
        }

        public Set<Integer> getCrashedVINs() {
            return this.crashedVINs;
        }

        public Set<Integer> getSpawnedVINs() {
            return spawnedVINs;
        }
    }

    private void fireStateChanged() {
        for (SimulatorStateChangeListener stateChangeListener : stateChangeListeners) {
            stateChangeListener.stateChanged(this);
        }
    }

    public void registerStateChangeListener(SimulatorStateChangeListener listener) {
        this.stateChangeListeners.add(listener);
    }

    public void registerSmallStepStatisticsListener(SmallScaleStatisticsEventListener listener){
        statisticsEventListeners.add(listener);
    }

    public Set<Integer> getVinsOfStoppedCars(double stoppingThreshold){
        return vinToVehicles
                .values()
                .stream()
                .filter(v -> v.getVelocity() < stoppingThreshold)
                .map(VehicleSimView::getVIN)
                .collect(Collectors.toSet());
    }

    private void updateStoppageTimes(double timeStepSize){
        Set<Integer> stoppedCars = getVinsOfStoppedCars(0.01);
        stoppedCars.forEach(vin -> stoppageTimeOfAllCars.put(vin, stoppageTimeOfAllCars.getOrDefault(vin, 0d) + timeStepSize));
    }

    public MOEASimStepResult step(double timeStepSize) {
        //distribute target speeds

        letDriversAct();

        Set<Integer> carsInsideIntersection = carsInsideInnerIntersection();
        moveVehicles(timeStepSize);
        carsInsideIntersection.removeAll(carsInsideInnerIntersection()); //Remove cars still in the intersection
        carsOutOfInnerIntersection.addAll(carsInsideIntersection);
        //fitness measurements (crashes throughput etc)

        List<Integer> completedVINs = cleanUpCompletedVehicles();
        currentTime += timeStepSize;
        Set<Integer> spawnedVINs;
        fireStateChanged();

        if (spawnVehicles)
            spawnedVINs = spawnVehicles(timeStepSize);
        else spawnedVINs = new HashSet<Integer>();
        updateStoppageTimes(timeStepSize);

        timeLeft -= timeStepSize;
        return new MOEASimStepResult(completedVINs, calculateCrashedVINs(), spawnedVINs);
    }

    private Set<Integer> carsInsideInnerIntersection() {
        Set<Integer> result = new HashSet<Integer>();
        for (VehicleSimView v: vinToVehicles.values()){
            if(v.getShape().intersects(intersection.getBoundingBox()))
                result.add(v.getVIN());
        }
        return result;
    }

    public Set<Shape> getSpawnPointZones() {
        Set<Shape> zones = new HashSet<Shape>();
        for (SpawnPoint spawnPoint : basicMap.getSpawnPoints()) {
            Rectangle2D noVehicleZone = spawnPoint.getNoVehicleZone();
            zones.add(noVehicleZone);
        }
        return zones;

    }

    private boolean vehicleInSpawnZone(VehicleSimView vehicle) {
        for (Shape zone : getSpawnPointZones()) {
            if (zone.intersects(vehicle.getShape().getBounds2D())) {
                return true;
            }
        }
        return false;
    }

    public Set<Integer> calculateCrashedVINs() {
        List<VehicleSimView> vehicles = new ArrayList<VehicleSimView>(vinToVehicles.values());
        Set<Integer> collisions = new HashSet<Integer>();
        for (int i = 0; i < vehicles.size(); i++) {
            for (int j = i + 1; j < vehicles.size(); j++) {
                VehicleSimView vehicle1 = vehicles.get(i);
                VehicleSimView vehicle2 = vehicles.get(j);
                if(SimUtils.distance(vehicle1,vehicle2) > 10)
                    continue;

                Area a1 = new Area(vehicle1.getShape());
                a1.intersect(new Area(vehicle2.getShape()));
                if (!a1.isEmpty() && !vehicleInSpawnZone(vehicle1)) {
                    collisions.add(vehicle1.getVIN());
                    collisions.add(vehicle2.getVIN());
                    //statisticsEventListeners.forEach((listener) -> listener.vehiclesCollide(this.currentTime, resolveState(vehicle1), resolveState(vehicle2)));
                }
            }
        }
        return collisions;
    }

    public void removeAllVehicles() {
        registry = new ThreadSafeVinRegistry(); //all the cars are gone. no need for old vins
        vinToVehicles.clear();
    }

    public BasicMap getMap() {
        return basicMap;
    }

    public double getSimulationTime() {
        return currentTime;
    }

    public int getNumCompletedVehicles() {
        return numOfCompletedVehicles;
    }

    public double getAvgBitsTransmittedByCompletedVehicles() {
        return 0;
    }

    public double getAvgBitsReceivedByCompletedVehicles() {
        return 0;
    }

    /**
     * {@inheritDoc}
     */
    public synchronized Set<VehicleSimView> getActiveVehicles() {
        return new HashSet<VehicleSimView>(vinToVehicles.values());
    }

    /**
     * {@inheritDoc}
     */
    public VehicleSimView getActiveVehicle(int vin) {
        return vinToVehicles.get(vin);
    }


    public void addProxyVehicle(ProxyVehicleSimView proxyVehicleSimView) {

    }

    /**
     * Set the target speed of all the vehicles to the same speed.
     *
     * @param speed
     */
    public void setAllVehicleSpeeds(double speed) {
        for (VehicleSimView vehicle : vinToVehicles.values()) {
            vehicle.setTargetVelocityWithMaxAccel(speed);
        }
    }

    public void setSpawnVehicles(boolean spawnVehicles) {
        this.spawnVehicles = spawnVehicles;
    }


    /**
     * returns a list of the vehicles sorted by their VIN. This ensures that for every state of the simulator we can
     * access the vehicles in the same order
     *
     * @return
     */
    public List<VehicleSimView> getSortedVehicles() {
        class VehicleComparator implements Comparator<VehicleSimView> {
            public int compare(VehicleSimView o1, VehicleSimView o2) {
                return o1.getVIN() - o2.getVIN();
            }
        }

        List<VehicleSimView> vehicles = new ArrayList<VehicleSimView>();
        vehicles.addAll(vinToVehicles.values());
        Collections.sort(vehicles, new VehicleComparator());
        return vehicles;
    }

    public void setSpecificVehicleSpeeds(double[] speeds) {
        List<VehicleSimView> vehicles = getSortedVehicles();
        if (vehicles.size() != speeds.length)
            throw new IllegalArgumentException("FFS, speed vector and vehicle list do not match....");

        for (int i = 0; i < speeds.length; i++) {
            vehicles.get(i).setTargetVelocityWithMaxAccel(speeds[i]);
        }
    }

    private VehicleState resolveState(VehicleSimView vehicle){
        return new VehicleState(
                vehicle.getDriver().getCurrentLane().getId(),
                ((StupidDriver)vehicle.getDriver()).departureLane.getId(),
                vehicle.getHeading(),
                ((BasicVehicle) vehicle).getSteeringAngle(),
                vehicle.getVelocity(),
                vehicle.getAcceleration(),
                vehicle.getVIN(),
                vehicle.getPosition().getX(),
                vehicle.getPosition().getY(),
                vehicle.getSpec().getName()
        );
    }

    /**
     * THis method may be solved by using a cast to BAsicVehicle. Will look into that
     *
     * @return
     */
    public Set<VehicleState> getAllVehicleStates() {
        Set<VehicleState> result = new HashSet<VehicleState>();

        for (VehicleSimView vehicle : vinToVehicles.values()) {

            result.add(resolveState(vehicle));
        }
        return result;
    }

    private void addVehicleToWorld(final VehicleSimView vehicle){
        vinToVehicles.put(vehicle.getVIN(), vehicle);
    }

    public void spawnVehiclesFromVehicleStates(Set<VehicleState> vehicleStates) {
        for (VehicleState state : vehicleStates) {
            VehicleSimView vehicle = spawnVehicleFromState(state);
            registry.registerVehicleWithGivenVin(vehicle);
            addVehicleToWorld(vehicle);
        }
    }

    private VehicleSimView spawnVehicleFromState(VehicleState vehicleState) {
        VehicleSpec spec = VehicleSpecDatabase.getVehicleSpecByName(vehicleState.carType);
        Lane currentLane = basicMap.getLaneRegistry().get(vehicleState.startLane);
        Lane goalLane = basicMap.getLaneRegistry().get(vehicleState.endLane);
        double initVelocity = vehicleState.velocity;
        // Obtain a Vehicle
        AutoVehicleSimView vehicle =
                new BasicAutoVehicle(spec,
                        new Point2D.Double(vehicleState.x, vehicleState.y),
                        vehicleState.heading,
                        vehicleState.steeringAngle,
                        initVelocity, // velocity
                        initVelocity,  // target velocity
                        vehicleState.acceleration,
                        0); //setting the clock in the car. Not needed if we are not doing FCFS
        // Set the driver
        AutoDriver driver = new StupidDriver(vehicle, currentLane, goalLane, basicMap, SimUtils.resolveDirection(currentLane.getId(), goalLane.getId()));
        driver.setCurrentLane(currentLane);

        vehicle.setDriver(driver);
        vehicle.setVIN(vehicleState.VIN);

        return vehicle;
    }

    //////////////////////////////////
    //COPIED FROM AIMSIM
    //////////////////////////////////

    /**
     * Whether a spawn point can spawn any vehicle
     *
     * @param spawnPoint the spawn point
     * @return Whether the spawn point can spawn any vehicle
     */
    private boolean canSpawnVehicle(SpawnPoint spawnPoint) {
        Rectangle2D noVehicleZone = spawnPoint.getNoVehicleZone();

        for (VehicleSimView vehicle : vinToVehicles.values()) {
            if (vehicle.getShape().intersects(noVehicleZone)) {
                return false;
            }
        }
        return true;
    }

    /**
     * Allow each driver to act.
     */
    private void letDriversAct() {
        for (VehicleSimView vehicle : vinToVehicles.values()) {
            vehicle.getDriver().act();
        }
    }

    /**
     * Spawn vehicles.
     *
     * @param timeStepSize the time step
     */
    private Set<Integer> spawnVehicles(double timeStepSize) {
        Set<Integer> spawnedVINs = new HashSet<Integer>();
        for (SpawnPoint spawnPoint : basicMap.getSpawnPoints()) {
            List<SpawnPoint.SpawnSpec> spawnSpecs = spawnPoint.act(timeStepSize);
            if (!spawnSpecs.isEmpty()) {
                if (canSpawnVehicle(spawnPoint)) {
                    for (SpawnPoint.SpawnSpec spawnSpec : spawnSpecs) {
                        VehicleSimView vehicle = makeVehicle(spawnPoint, spawnSpec);
                        registry.registerVehicle(vehicle);
                        addVehicleToWorld(vehicle);
                        statisticsEventListeners.forEach((listener) -> listener.vehicleSpawned(this.currentTime, vehicle.getVIN()));
                        spawnedVINs.add(vehicle.getVIN());
                        break; // only handle the first spawn vehicle
                    }
                } // else ignore the spawnSpecs and do nothing
            }
        }
        return spawnedVINs;
    }

    /**
     * Create a vehicle at a spawn point.
     *
     * @param spawnPoint the spawn point
     * @param spawnSpec  the spawn specification
     * @return the vehicle
     */
    private VehicleSimView makeVehicle(SpawnPoint spawnPoint,
                                       SpawnPoint.SpawnSpec spawnSpec) {
        VehicleSpec spec = spawnSpec.getVehicleSpec();
        Lane lane = spawnPoint.getLane();
        // Now just take the minimum of the max velocity of the vehicle, and
        // the speed limit in the lane
        double spawnZoneLength = 28.0;
        double initVelocity = Math.min(spec.getMaxVelocity(), lane.getSpeedLimit());

        //initVelocity = Math.min(initVelocity, initVelocity * timeLeft / 28d);
        if(timeLeft > 0 && 20.5 / (timeLeft) > 0){
            initVelocity = Math.min(20.5 / timeLeft, initVelocity);
        }
        //System.out.println(initVelocity);
        double stoppingDistanceNeeded = Math.pow(initVelocity, 2) / (2* spec.getMaxDeceleration());
        //System.out.println(stoppingDistanceNeeded);
        // Obtain a Vehicle
        AutoVehicleSimView vehicle =
                new BasicAutoVehicle(spec,
                        spawnPoint.getPosition(),
                        spawnPoint.getHeading(),
                        spawnPoint.getSteeringAngle(),
                        initVelocity, // velocity
                        initVelocity,  // target velocity
                        spawnPoint.getAcceleration(),
                        spawnSpec.getSpawnTime());
        // Set the driver
        int destinationIndexLane = spawnSpec.getDestinationRoad().getIndexLane().getId();
        int startLaneId = lane.getId();

        //If its a uturn, lets just go straight instead.
        if(SimUtils.checkIsUturn(startLaneId, destinationIndexLane))
            destinationIndexLane = startLaneId;

        SimUtils.Direction dir = SimUtils.resolveDirection(startLaneId,destinationIndexLane);
        int correctStartLaneId = startLaneId;
        int correctDestinationLane = SimUtils.resolveCorrectLaneIDfromCurrentLaneIDAndDirection(startLaneId, dir);
        //The direction may have changed if the goal lane conflicted with the starting lane
        dir = SimUtils.resolveDirection(startLaneId, correctDestinationLane);

        if(dir == SimUtils.Direction.STRAIGHT){
            //Just stay in lane if we are going straight
            correctDestinationLane = startLaneId;
            correctStartLaneId = startLaneId;
        }

        AutoDriver driver = new StupidDriver(vehicle, basicMap.getLaneRegistry().get(correctStartLaneId), basicMap.getLaneRegistry().get(correctDestinationLane), basicMap,dir);
        //AutoDriver driver = new AutoDriver(vehicle, basicMap);
        driver.setCurrentLane(lane);
        driver.setSpawnPoint(spawnPoint);
        driver.setDestination(spawnSpec.getDestinationRoad());
        vehicle.setDriver(driver);

        return vehicle;
    }

    private double getKineticEnergy(VehicleSimView vehicle) {
        return 0.5 * 1 * Math.pow(vehicle.getVelocity(), 2);
    }


    /**
     * gives total kinetic energy in the simulator, in KJ per car
     * Assumes the weight of each car to be 1000KG
     *
     * @return kinetic energy per car, indentified by VIN
     */
    public Map<Integer, Double> getTotalKineticEnergy() {
        Map<Integer, Double> ek = new HashMap<Integer, Double>();
        for (VehicleSimView vehicle :
                vinToVehicles.values()) {
            ek.put(vehicle.getVIN(), getKineticEnergy(vehicle));
        }
        return ek;
    }

    /**
     * Move all the vehicles.
     *
     * @param timeStep the time step
     */
    private void moveVehicles(double timeStep) {
        distanceTraveled = 0;
        for (VehicleSimView vehicle : vinToVehicles.values()) {
            Point2D p1 = vehicle.getPosition();
            vehicle.move(timeStep);
            Point2D p2 = vehicle.getPosition();
            distanceTraveled += p1.distance(p2);
            for (DataCollectionLine line : basicMap.getDataCollectionLines()) {
                line.intersect(vehicle, currentTime, p1, p2);
            }
            if (Debug.isPrintVehicleStateOfVIN(vehicle.getVIN())) {
                vehicle.printState();
            }
        }
    }

    public double getDistanceTraveled() {
        return distanceTraveled;
    }

    /**
     * Remove all completed vehicles.
     *
     * @return the VINs of the completed vehicles
     */
    private List<Integer> cleanUpCompletedVehicles() {
        List<Integer> completedVINs = new ArrayList<>();

        Rectangle2D mapBoundary = basicMap.getDimensions();

        List<Integer> removedVINs = new ArrayList<Integer>(vinToVehicles.size());
        for (int vin : vinToVehicles.keySet()) {
            VehicleSimView v = vinToVehicles.get(vin);
            // If the vehicle is no longer in the layout
            if (!v.getShape().intersects(mapBoundary)) {

                removedVINs.add(vin);
            }
        }
        // Remove the marked vehicles
        for (int vin : removedVINs) {
            vinToVehicles.remove(vin);
            statisticsEventListeners.forEach((listener) -> listener.vehicleRemoved(this.currentTime, vin));
            completedVINs.add(vin);
            numOfCompletedVehicles++;
        }

        return completedVINs;
    }

    public Set<Integer> getCarsOutOfInnerIntersection() {
        return carsOutOfInnerIntersection;
    }
}
