package jossihawk.simulator.sim;

import aim4.sim.setup.BasicSimSetup;
import aim4.vehicle.VehicleSimView;
import jossihawk.simulator.AIMSimulator;
import jossihawk.simulator.AIMStats;
import jossihawk.simulator.SimulationProperties;
import jossihawk.simulator.SimulationState;
import jossihawk.simulator.sim.stats.StatCounter;
import jossihawk.simulator.sim.stats.VehicleStatistics;
import jossihawk.simulator.sim_view.Display;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * This class encapsulates the MOEASimulator, and corresponds with the AIMSimulator contract.
 * The goal of the class is to hide the complexity of the AIM framework from the MOEA algorithm.
 */
public class SimulatorController extends AIMSimulator {

    private final StatCounter smallStepStatCounter;

    private final MOEASimulator simulator;
    private final MOEASimSetup setup;
    private double distanceTraveledInLastTimeStep, kineticEnergyLostInLastTimeStep;
    private int carsEvacuatedFromInnerIntersectionInLastTimeStep;
    private Set<Integer> totalChrashedVINs;


    private double simulatorDT = 0.02; //Just using the same dt as the AIM guys.

    private final SimulationProperties properties;
    private int numberOfCollisions, numberOfCompletedVehicles;
    private Map<Integer, VehicleStatistics> vehicleStatisticsMap;
    private boolean mainSimulator;
    private boolean headless;

    public SimulatorController(SimulationState simulationState){
        SimulationProperties simulationProperties = simulationState.getSimulationProperties();
        BasicSimSetup simSetup
                = new BasicSimSetup(1, // columns
                1, // rows
                6, // lane width
                simulationProperties.getSpeedLimit(), // speed limit
                simulationProperties.getNumberOfLanes(), // lanes per road
                1.5, // median size
                150, // distance between
                0.0, //Irrelevant
                1.0 // stop distance before intersection
        );
        this.properties = simulationProperties;
        setup = new MOEASimSetup(simSetup, properties);
        simulator = setup.getSimulator();
        if( simulationState.getVehicleStates() != null)
            simulator.spawnVehiclesFromVehicleStates(simulationState.getVehicleStates());

        smallStepStatCounter = new StatCounter(simulator);
        totalChrashedVINs = new HashSet<>();

    }

    public void setAsMainSimulator(boolean headless){
        mainSimulator = true;
        this.headless = headless;
        vehicleStatisticsMap = new HashMap<>();
        if(!headless)
            new Display(setup, simulator);
    }

    @Override
    public double getSimulationTime() {
        return simulator.getSimulationTime();
    }

    @Override
    public double getTotalDistanceTraveledInLastTimeStep() {
        return distanceTraveledInLastTimeStep;
    }

    @Override
    public double getKineticEnergyLostInLastTimeStep() {
        return kineticEnergyLostInLastTimeStep;
    }

    @Override
    public double getCarsEvacuatedFromInnerIntersectionInLastTimeStep() {
        return carsEvacuatedFromInnerIntersectionInLastTimeStep;
    }

    public double getStoppageTimeOfStoppedCars(){
        return simulator.getStoppageTimeOfStoppedCars();
    }

    private MOEASimulator.MOEASimStepResult doSimulatorStep(double dt){
        MOEASimulator.MOEASimStepResult result = simulator.step(dt);
        if(mainSimulator)
            updateVehicleStatisticsMap(result);
        distanceTraveledInLastTimeStep += simulator.getDistanceTraveled();
        return result;
    }

    private void updateVehicleStatisticsMap(MOEASimulator.MOEASimStepResult result) {
        for (VehicleSimView vehicle : simulator.getActiveVehicles()) {
            if(!vehicleStatisticsMap.containsKey(vehicle.getVIN()))
                vehicleStatisticsMap.put(vehicle.getVIN(), new VehicleStatistics(simulator.getSimulationTime()));
        }
        for (Integer completedVIN : result.completedVINs) {
            vehicleStatisticsMap.get(completedVIN).setEvacuationTime(simulator.getSimulationTime());
        }
    }

    private MOEASimulator.MOEASimStepResult doUnderlyingStepInRealTime(){
        long start = System.currentTimeMillis();
        MOEASimulator.MOEASimStepResult result = doSimulatorStep(simulatorDT);
        long timeUsed = System.currentTimeMillis() - start;

        if (timeUsed < simulatorDT * 1000){
            try {
                Thread.sleep( (long) (simulatorDT * 1000) - timeUsed);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }else{
            System.err.println("The real time simulator is having problems keeping up!");
        }
        return result;
    }

    private void step(double[] speeds, double timeToSimulate) {
        distanceTraveledInLastTimeStep = 0;
        carsEvacuatedFromInnerIntersectionInLastTimeStep = 0;

        int carsAlreadyEvacuatedFromInnerTimeStep = simulator.getCarsOutOfInnerIntersection().size();
        Map<Integer, Double> energyLevelsBeforeTimeStep = simulator.getTotalKineticEnergy();
        Set<Integer> crashedVINs = new HashSet<>();
        if (timeToSimulate < simulatorDT){
            throw new IllegalArgumentException(String.format("The given time is less than one timestep in the underlying simulator. Must exceed %d", simulatorDT));
        }

        simulator.setSpecificVehicleSpeeds(speeds);
        simulator.timeLeft = timeToSimulate;
        //calculate how many simulation steps we need in order to meet the required simulation time
        int simulationSteps = (int) (timeToSimulate / simulatorDT);
        for (int i = 0; i < simulationSteps; i++){
            if(mainSimulator && !headless)
                crashedVINs.addAll(doUnderlyingStepInRealTime().crashedVINs);
            else
                crashedVINs.addAll(doSimulatorStep(simulatorDT).crashedVINs);
        }
        Map<Integer, Double> energyLevelsAfterTimeStep = simulator.getTotalKineticEnergy();

        carsEvacuatedFromInnerIntersectionInLastTimeStep = simulator.getCarsOutOfInnerIntersection().size() - carsAlreadyEvacuatedFromInnerTimeStep;
        kineticEnergyLostInLastTimeStep = calculateEnergyLostToBreaking(energyLevelsBeforeTimeStep, energyLevelsAfterTimeStep);
        numberOfCollisions = crashedVINs.size();
        totalChrashedVINs.addAll(crashedVINs);
        numberOfCompletedVehicles = simulator.getNumCompletedVehicles();


    }

    private double calculateEnergyLostToBreaking(Map<Integer, Double> energyBefore, Map<Integer, Double> energyAfter){
        double ekDiff = 0;

        for (Integer vin : energyBefore.keySet()) {
            if(energyAfter.containsKey(vin) && energyBefore.get(vin) > energyAfter.get(vin)) {
                ekDiff += energyBefore.get(vin) - energyAfter.get(vin);
                if(mainSimulator){
                    vehicleStatisticsMap.get(vin).setSpentEk(vehicleStatisticsMap.get(vin).getSpentEk() + ekDiff);
                }
            }
        }

        return ekDiff;
    }

    @Override
    public int getNumberOfActiveCars() {
        return simulator.getActiveVehicles().size();
    }

    @Override
    public void step(double[] speeds) {
        step(speeds, properties.getAlgorithmTimeStepSize());
    }

    @Override
    public void evaluate(double[] speeds) {
        step(speeds, properties.getEvaluationSimulationTime());
    }

    @Override
    public AIMStats getTotalStats() {
        return new AIMStats(vehicleStatisticsMap, totalChrashedVINs.size());
    }

    @Override
    public int getNumberOfCollisions() {
        return numberOfCollisions;
    }

    public MOEASimulator getSimulator() {
        return simulator;
    }

    @Override
    public int getThroughput() {
        return numberOfCompletedVehicles;
    }

    @Override
    public SimulationState getState() {
        Set<VehicleState> vehicleStates = simulator.getAllVehicleStates();
        SimulationProperties props = new SimulationProperties(this.properties, vehicleStates.size());
        SimulationState state = new SimulationState(vehicleStates, props);
        return state;
    }
    public void setSpawnVehicles(boolean spawnVehicles) {
        simulator.setSpawnVehicles(spawnVehicles);
    }

    public void removeAllVehicles(){
        simulator.removeAllVehicles();
    }

    public void reinitializeFromState(SimulationState state){
        simulator.removeAllVehicles();
        simulator.spawnVehiclesFromVehicleStates(state.getVehicleStates());
    }
}
