package jossihawk.simulator.sim;

import aim4.driver.AutoDriver;
import aim4.driver.DriverUtil;
import aim4.map.BasicMap;
import aim4.map.lane.Lane;
import aim4.vehicle.AutoVehicleDriverView;

import java.awt.geom.Point2D;

/**
 * Largely an adaption of the simple driver class from AIM4. However, changes that were impossible by overriding had to
 * be made because the needed methods had the wrong privacy declarations. Needed to make the "stupid" autonomous vehicles
 * behave as slaves of the IM, and not as individual thinkers. 
 */
public class StupidDriver extends AutoDriver {



    /////////////////////////////////
    // PRIVATE FIELDS
    /////////////////////////////////

    /** The vehicle this driver will control */
    private AutoVehicleDriverView vehicle;

    /** The Lane in which the vehicle should exit the intersection. */
    private int laneIdWeWereToldToUse;
    public final Lane departureLane;
    BasicMap currentMap;
    SimUtils.Direction direction; //Which way are we turning?


    /////////////////////////////////
    // CLASS CONSTRUCTORS
    /////////////////////////////////

    /**
     * Construct a new CrashTestDummy to pilot the simulated vehicle across
     * an intersection.
     *
     * @param vehicle       the simulated vehicle to pilot
     * @param arrivalLane   the Lane in which the vehicle should enter the
     *                      intersection
     * @param departureLane the Lane in which the vehicle should depart the
     *                      intersection
     */
    public StupidDriver(AutoVehicleDriverView vehicle,
                          Lane arrivalLane, Lane departureLane, BasicMap map, SimUtils.Direction direction) {
        super(vehicle,map);
        currentMap = map;
        this.vehicle = vehicle;
        setCurrentLane(arrivalLane);
        this.departureLane = departureLane;
        this.direction = direction;
        laneIdWeWereToldToUse = arrivalLane.getId();
    }

    /////////////////////////////////
    // PUBLIC METHODS
    /////////////////////////////////

    /**
     * Take steering actions to guide a test vehicle through a simulated
     * traversal of the intersection.
     */
    @Override
    public void act() {
        //We don't need or want a coordinator
        //super.act();
        // If we're not already in the departure lane
        if(getCurrentLane() != departureLane) {
            //If we are changing lane internally in a road
            if(getCurrentLane().getId() != laneIdWeWereToldToUse)
                setCurrentLane(currentMap.getLaneRegistry().get(laneIdWeWereToldToUse));
            // If we're changing to a different Road
            else if(currentMap.getRoad(getCurrentLane()) != currentMap.getRoad(departureLane)) {
                // If we're close enough...
                if(departureLane.nearestDistance(getVehicle().gaugePosition()) <
                        calculateTraversingLaneChangeDistance()) {
                    // Change to it
                    setCurrentLane(departureLane);
                }
            } else {
                // Otherwise, we're changing to the same Road, so we need a different
                // criterion... in this case none
                setCurrentLane(departureLane);
            }
        }
        // Use the basic lane-following behavior
        followCurrentLane();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public AutoVehicleDriverView getVehicle() {
        return vehicle;
    }

    /**
     * Turn the wheels to follow the current lane, using the
     * <code>DEFAULT_LEAD_TIME</code>. This involves first projecting
     * the Vehicle's current position onto the lane, and then projecting
     * forward by a distance equal to the Vehicle's velocity multiplied
     * by the lead time.
     */
    private void followCurrentLane() {
        followCurrentLane(DriverUtil.DEFAULT_LEAD_TIME);
    }


    /**
     * Turn the wheels to follow the current lane, using the given lead time.
     * This involves first projecting the Vehicle's current position onto the
     * lane, and then projecting forward by a distance equal to the Vehicle's
     * velocity multiplied by the lead time.
     *
     * @param leadTime the lead time to use
     */
    private void followCurrentLane(double leadTime) {
        double leadDist = leadTime * vehicle.gaugeVelocity() +
                DriverUtil.MIN_LEAD_DIST;
        Point2D aimPoint;
        double remaining = getCurrentLane().
                remainingDistanceAlongLane(vehicle.gaugePosition());
        // If there's not enough room in this Lane and there is a Lane that this
        // Lane leads into, use the next Lane
        if((leadDist > remaining) && (getCurrentLane().hasNextLane())) {
            // First make sure we shouldn't transfer to the next lane
            if(remaining <= 0) {
                // Switch to the next Lane
                setCurrentLane(getCurrentLane().getNextLane());
                // currentLane = currentLane.getNextLane();
                // And do this over
                followCurrentLane(leadTime);
                return;
            }
            // Use what's left over after this Lane to go into the next one.
            aimPoint = getCurrentLane().getNextLane().getLeadPoint(
                    getCurrentLane().getNextLane().getStartPoint(),
                    leadDist - remaining);
            // Indicate that this is the point for which we are aiming
//      addDebugPoint(new DebugPoint(aimPoint,
//                                   vehicle.gaugePointBetweenFrontWheels(),
//                                   "next lane", Color.RED));
        } else { // Otherwise, use the current Lane
            aimPoint = getCurrentLane().getLeadPoint(
                    vehicle.gaugePosition(), leadDist);
            // Indicate that this is the point for which we are aiming
//      addDebugPoint(new DebugPoint(aimPoint,
//                                   vehicle.gaugePointBetweenFrontWheels(),
//                                   "lane", Color.PINK.brighter()));
        }
        turnTowardPoint(aimPoint);
    }


    /**
     * Turn the wheels toward a given Point.
     *
     * @param p the Point toward which to turn the wheels
     */
    private void turnTowardPoint(Point2D p) {
        vehicle.turnTowardPoint(p);
    }


    /**
     * Determine the distance at which to change lanes during the
     * traversing inside an intersection.  Currently just
     * multiplies the TRAVERSING_LANE_CHANGE_LEAD_TIME by the
     * reading from the Vehicle's speedometer.
     *
     * @return the distance at which the Driver should turn the Vehicle into
     *         the lane to which it is changing
     */
    public double calculateTraversingLaneChangeDistance() {
        if(direction != SimUtils.Direction.LEFT)//Right hand turns are a lot tighter
            return 1.1;
        return 8;//V2IPilot.TRAVERSING_LANE_CHANGE_LEAD_TIME * vehicle.gaugeVelocity()*0.8;

    }


}
