package jossihawk.utils;

import jossihawk.simulator.AIMStats;
import org.jfree.data.xy.XYDataItem;
import org.jfree.data.xy.XYSeries;

import javax.xml.bind.JAXBException;
import java.io.*;
import java.util.*;
import java.util.stream.Collectors;

public class ExperimentAnalyser {

    private final File analyticsDirectory;

    public enum Objective {
        MEANEVACUATIONTIME, KINETICENERGYLOST, THROUGHPUT, CRASHEDVEHICLES;
    }

    public enum TrafficDensity {
        HIGH, MEDIUM, LOW, ALL;
    }

    private List<ScenarioResult> scenarioResults;
    private File experimentDirectory;

    public ExperimentAnalyser(File experimentDirectory) throws IOException, JAXBException {
        File[] scenarios = experimentDirectory.listFiles(pathname -> {
            return new File(pathname + File.separator + "results" + File.separator + "globalStats.xml").exists();
        });
        this.experimentDirectory = experimentDirectory;

        if (scenarios == null)

        {
            throw new RuntimeException("Aborting! No experiment directory: " + experimentDirectory.getAbsolutePath());
        }
        scenarioResults = new ArrayList<>();
        for (
                File scenario
                : scenarios)

        {

            Properties runProperties = readPropertiesFile(new File(scenario, scenario.getName() + ".properties"));
            double tMain = Double.parseDouble(runProperties.getProperty("algorithmTimeStepSize"));
            double tEval = Double.parseDouble(runProperties.getProperty("evaluationSimulationTime"));
            AIMStats aimStats = AIMStats.readFromFile(new File(scenario, "results" + File.separator + "globalStats.xml"));

            //State vs nostate
            File additionalPropertiesFile = new File(scenario, "additional.properties");
            Properties additionalProperties;
            if (additionalPropertiesFile.exists()) {
                additionalProperties = readPropertiesFile(additionalPropertiesFile);
                String trafficDensity = additionalProperties.getProperty("trafficDensity");
                String state = additionalProperties.getProperty("state");
                scenarioResults.add(new ScenarioResult(trafficDensity, state, tMain, tEval, aimStats, scenario));
            } else {
                scenarioResults.add(new ScenarioResult(tMain, tEval, aimStats, scenario));
            }


        }
        analyticsDirectory = new File(experimentDirectory, "analytics");
        if (!analyticsDirectory.exists())
            analyticsDirectory.mkdir();
    }

    public void writeGraph(TrafficDensity trafficDensity, Objective objective) throws FileNotFoundException, UnsupportedEncodingException {
        XYSeries series = createAVGXYseries(trafficDensity,objective);
        String filename = trafficDensity.toString() + "-" + objective.toString() + ".csv";
        File graph = new File(analyticsDirectory, filename);
        PrintWriter writer = new PrintWriter(graph, "UTF-8");
        writer.println("steps,objective");
        for(int i = 0; i < series.getItemCount(); i++) {
            XYDataItem dataItem = series.getDataItem(i);
            writer.println(dataItem.getX() + "," + dataItem.getYValue());
        }
        writer.close();
    }

    public void writeGraphs(TrafficDensity trafficDensity, Objective[] objectives) throws FileNotFoundException, UnsupportedEncodingException {
        for (Objective objective : objectives) {
            writeGraph(trafficDensity, objective);
        }
    }

    private List<ScenarioResult> filterOnTrafficDensity(List<ScenarioResult> scenarioResults, TrafficDensity trafficDensity) {
        return scenarioResults.stream().filter(item -> trafficDensity == TrafficDensity.ALL || item.getTrafficDensity() == trafficDensity).collect(Collectors.toList());
    }

    private XYSeries createXYSeries(TrafficDensity trafficDensity, Objective objective) {
        XYSeries series = new XYSeries(objective.toString());
        for (ScenarioResult scenarioResult : filterOnTrafficDensity(scenarioResults, trafficDensity)) {
            if (scenarioResult.gettMain() < 250)
                series.add(scenarioResult.gettMain(), scenarioResult.getObjective(objective));
        }
        return series;
    }

    private XYSeries createAVGXYseries(TrafficDensity trafficDensity, Objective objective) {
        Map<Double, List<Double>> timeStepMap = new HashMap<>();
        for (ScenarioResult scenarioResult : filterOnTrafficDensity(scenarioResults, trafficDensity)) {
            double timeStep = scenarioResult.gettMain();
            double objectiveValue = scenarioResult.getObjective(objective);
            if (timeStepMap.containsKey(timeStep))
                timeStepMap.get(timeStep).add(objectiveValue);
            else {
                List<Double> timeStepObjectives = new ArrayList<>();
                timeStepObjectives.add(objectiveValue);
                timeStepMap.put(timeStep, timeStepObjectives);
            }
        }
        XYSeries dataPoints = new XYSeries(objective.toString());
        for (Double timeStep : timeStepMap.keySet()) {
            if (timeStep < 203) {
                Double avg = timeStepMap.get(timeStep).stream().mapToDouble(i -> i).average().getAsDouble();
                dataPoints.add(timeStep, avg);
            }
        }
        return dataPoints;
    }

    public List<File> getCrashedScenarioFiles(TrafficDensity trafficDensity, double timeStepSize) {
        List<File> chrashedScenarioFiles = new ArrayList<>();
        for (ScenarioResult scenarioResult : filterOnTrafficDensity(scenarioResults, trafficDensity)) {
            if (timeStepSize == scenarioResult.gettMain() && scenarioResult.getNofChrashedVehicles() > 0) {
                chrashedScenarioFiles.add(scenarioResult.getScenarioFile());
            }
        }
        return chrashedScenarioFiles;
    }

    public void showGraph(TrafficDensity trafficDensity, Objective objective, boolean average) {
        List<XYSeries> xySeriesList = new ArrayList<>();
        XYSeries xySeries;
        if (average)
            xySeries = createAVGXYseries(trafficDensity, objective);
        else xySeries = createXYSeries(trafficDensity, objective);
        xySeriesList.add(xySeries);
        new ChartDrawer(xySeriesList, objective.toString());
    }

    public void showGraph(TrafficDensity trafficDensity, Objective[] objectives, boolean average) {
        List<XYSeries> xySeriesList = new ArrayList<>();
        for (Objective objective : objectives) {
            XYSeries xySeries;
            if (average)
                xySeries = createAVGXYseries(trafficDensity, objective);
            else xySeries = createXYSeries(trafficDensity, objective);
            xySeriesList.add(xySeries);
        }
        new ChartDrawer(xySeriesList, "Collection of objectives graph");
    }

    private XYSeries createAVGXYseries(TrafficDensity trafficDensity, Objective objective, double tMain) {
        Map<Double, List<Double>> timeStepMap = new HashMap<>();
        for (ScenarioResult scenarioResult : scenarioResults) {
            if (scenarioResult.gettMain() == tMain) {
                double timeStep = scenarioResult.gettEval()-tMain;
                double objectiveValue = scenarioResult.getObjective(objective);
                if (timeStepMap.containsKey(timeStep))
                    timeStepMap.get(timeStep).add(objectiveValue);
                else {
                    List<Double> timeStepObjectives = new ArrayList<>();
                    timeStepObjectives.add(objectiveValue);
                    timeStepMap.put(timeStep, timeStepObjectives);
                }
            }
        }
        XYSeries dataPoints = new XYSeries(objective.toString() + "- tmain - " + tMain);
        for (Double timeStep : timeStepMap.keySet()) {
            Double avg = timeStepMap.get(timeStep).stream().mapToDouble(i -> i).average().getAsDouble();
            dataPoints.add(timeStep, avg);
        }
        return dataPoints;
    }

    public void showGraph(TrafficDensity trafficDensity, Objective[] objectives, double tMain) {
        for (Objective objective : objectives) {
            List<XYSeries> xySeriesList = new ArrayList<>();
            XYSeries xySeries;
            xySeries = createAVGXYseries(trafficDensity, objective, tMain);
            xySeriesList.add(xySeries);
            new ChartDrawer(xySeriesList, objective.toString() + "  : tmain - " + tMain);
        }
    }

    public void showAllExp1p2Graphs(TrafficDensity trafficDensity, Objective[] objectives) {
        Set<Double> tMains = new HashSet<>();
        for (ScenarioResult scenarioResult : scenarioResults) {
            tMains.add(scenarioResult.gettMain());
        }
        for (Double tMain : tMains) {
            showGraph(trafficDensity, objectives, tMain);
        }
    }


    public List<ScenarioResult> getScenarioResults() {
        return scenarioResults;
    }

    private Properties readPropertiesFile(File propertiesFile) throws IOException {
        FileInputStream fileInput = new FileInputStream(propertiesFile);
        Properties properties = new Properties();
        properties.load(fileInput);
        fileInput.close();
        return properties;
    }

    public class ScenarioResult {

        private double tEval;
        private String trafficDensity, state;
        private double tMain;
        private double meanEvacTime;
        private double totalSpentEk;
        private int throughput, nofChrashedVehicles;
        private File scenarioFile;

        public ScenarioResult(String trafficDensity, String state, double tMain, double tEval, AIMStats aimStats, File scenarioFile) {
            this.tMain = tMain;
            this.totalSpentEk = aimStats.getTotalSpentEk();
            this.meanEvacTime = aimStats.getMeanEvacTime();
            this.throughput = aimStats.getThroughput();
            this.nofChrashedVehicles = aimStats.getNofChrashedVehicles();
            this.scenarioFile = scenarioFile;
            this.tEval = tEval;

            this.trafficDensity = trafficDensity;
            this.state = state;
        }

        public ScenarioResult(double tMain, double tEval, AIMStats aimStats, File scenarioFile) {
            this.tMain = tMain;
            this.totalSpentEk = aimStats.getTotalSpentEk();
            this.meanEvacTime = aimStats.getMeanEvacTime();
            this.throughput = aimStats.getThroughput();
            this.nofChrashedVehicles = aimStats.getNofChrashedVehicles();
            this.scenarioFile = scenarioFile;
            this.tEval = tEval;
        }

        public double getObjective(Objective objective) {
            switch (objective) {
                case MEANEVACUATIONTIME:
                    return meanEvacTime;
                case KINETICENERGYLOST:
                    return totalSpentEk;
                case THROUGHPUT:
                    return throughput;
                case CRASHEDVEHICLES:
                    return nofChrashedVehicles;
                default:
                    throw new RuntimeException("Unknown objective: " + objective);
            }
        }

        public TrafficDensity getTrafficDensity() {
            switch (trafficDensity) {
                case "low-traffic":
                    return TrafficDensity.LOW;
                case "medium-traffic":
                    return TrafficDensity.MEDIUM;
                case "high-traffic":
                    return TrafficDensity.HIGH;
                default:
                    throw new RuntimeException("Didn't match any of the traffic-densities");
            }
        }

        public double gettMain() {
            return tMain;
        }

        public double getMeanEvacTime() {
            return meanEvacTime;
        }

        public double getTotalSpentEk() {
            return totalSpentEk;
        }

        public int getThroughput() {
            return throughput;
        }

        public int getNofChrashedVehicles() {
            return nofChrashedVehicles;
        }

        public String getState() {
            return state;
        }

        public File getScenarioFile() {
            return scenarioFile;
        }

        public double gettEval() {
            return tEval;
        }

        @Override
        public String toString() {
            return "ScenarioResult{" +
                    "trafficDensity='" + trafficDensity + '\'' +
                    ", state='" + state + '\'' +
                    ", tMain=" + tMain +
                    ", meanEvacTime=" + meanEvacTime +
                    ", totalSpentEk=" + totalSpentEk +
                    ", throughput=" + throughput +
                    ", nofChrashedVehicles=" + nofChrashedVehicles +
                    '}';
        }

    }

    public void displayPart2(TrafficDensity trafficDensity, Objective[] objectives) throws FileNotFoundException, UnsupportedEncodingException {
        Set<Double> tMains = new HashSet<>();
        for (ScenarioResult scenarioResult : scenarioResults) {
            tMains.add(scenarioResult.gettMain());
        }
        for (int i = 0; i < objectives.length; i++) {
            Objective objective = objectives[i];
            List<XYSeries> xySeriesList = new ArrayList<>();
            for (Double tMain : tMains) {
                xySeriesList.add(createAVGXYseries(trafficDensity, objective, tMain));
            }
            new ChartDrawer(xySeriesList, objective.toString());
            writePart2(xySeriesList, objective.toString());
        }
    }

    private void writePart2(List<XYSeries> xySeriesList, String key) throws FileNotFoundException, UnsupportedEncodingException {
        for (XYSeries series : xySeriesList) {
            File graph = new File(experimentDirectory.getAbsolutePath() + File.separator +  "analytics" + File.separator + series.getKey() +".csv");
            PrintWriter writer = new PrintWriter(graph, "UTF-8");
            writer.println("steps," + series.getKey());
            for (int i = 0; i < series.getItemCount(); i++) {
                XYDataItem dataItem = series.getDataItem(i);
                writer.println(dataItem.getXValue() + "," + dataItem.getYValue());
            }
            writer.close();
        }
    }

    public static void main(String[] args) throws IOException, JAXBException {
        ExperimentAnalyser experimentAnalyser = new ExperimentAnalyser(new File("experiments" + File.separator + "ex1p1_5runs"));
        boolean average = true;
        experimentAnalyser.showGraph(TrafficDensity.ALL, Objective.CRASHEDVEHICLES, average);
        experimentAnalyser.showGraph(TrafficDensity.ALL, Objective.KINETICENERGYLOST, average);
        experimentAnalyser.showGraph(TrafficDensity.ALL, Objective.MEANEVACUATIONTIME, average);
        experimentAnalyser.showGraph(TrafficDensity.ALL, Objective.THROUGHPUT, average);

//        experimentAnalyser.displayPart2(TrafficDensity.ALL, Objective.values());
        experimentAnalyser.writeGraphs(TrafficDensity.ALL, Objective.values());
    }
}
