package jossihawk.utils;

import jossihawk.simulator.AIMStats;
import jossihawk.simulator.SimulationProperties;

import javax.xml.bind.JAXBException;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

public class ExperimentFourAnalyzer {

    private Map<ExperimentFourKey, Map<Double, List<ExperimentFourResult>>> resultsMap;
    private List<File> experiments;

    public ExperimentFourAnalyzer(List<File> experiments) throws IOException, JAXBException {
        resultsMap = new HashMap<>();
        this.experiments = experiments;
        for (File experimentDir : experiments) {
            File[] resultFiles = experimentDir.listFiles(pathname -> {
                return new File(pathname + File.separator + "results" + File.separator + "globalStats.xml").exists();
            });

            for (int i = 0; i < resultFiles.length; i++) {
                File resultfile = new File(resultFiles[i].getAbsolutePath() + File.separator + "results" + File.separator + "globalStats.xml");
                AIMStats stats = AIMStats.readFromFile(resultfile);
                SimulationProperties props = new SimulationProperties(resultFiles[i].getAbsolutePath() + File.separator + resultFiles[i].getName() + ".properties");
                ExperimentFourResult experimentFourResult = new ExperimentFourResult(props, stats);
                double innerMapKey = experimentFourResult.getProperties().getSpawnRateNorth();
                if (resultsMap.containsKey(experimentFourResult.getKey())) {
                    if (resultsMap.get(experimentFourResult.getKey()).containsKey(innerMapKey)) {
                        resultsMap.get(experimentFourResult.getKey()).get(innerMapKey).add(experimentFourResult);
                    } else {
                        List<ExperimentFourResult> tempList = new ArrayList<>();
                        tempList.add(experimentFourResult);
                        resultsMap.get(experimentFourResult.getKey()).put(innerMapKey, tempList);
                    }
                } else {
                    List<ExperimentFourResult> tempList = new ArrayList<>();
                    tempList.add(experimentFourResult);
                    Map<Double, List<ExperimentFourResult>> innerMap = new HashMap<>();
                    innerMap.put(innerMapKey, tempList);
                    resultsMap.put(experimentFourResult.getKey(), innerMap);
                }
            }
        }
    }

    public void printRunningStatistics() {
        for (File experiment : experiments) {
            File[] finishedRuns = experiment.listFiles(pathname -> {
                return new File(pathname + File.separator + "results" + File.separator + "globalStats.xml").exists();
            });
            File[] currentRuns = experiment.listFiles(pathname -> {
                return !(new File(pathname + File.separator + "results" + File.separator + "globalStats.xml").exists()) && new File(pathname + File.separator + "results").exists();
            });
            File[] totalRuns = experiment.listFiles(pathname -> {
                return !pathname.getName().equals("analytics");
            });

            System.out.println("Experiment: " + experiment.getName());
            System.out.println("finished runs: " + finishedRuns.length);
            System.out.println("current runs: " + (currentRuns.length));
            System.out.println("total runs: " + totalRuns.length);

            for (int i = 0; i < currentRuns.length; i++) {
                File currentRun = currentRuns[i];
                System.out.println(currentRun);
            }


        }
    }

    public void printResults(boolean latex, String separator) {
        for (ExperimentFourKey experimentFourKey : resultsMap.keySet()) {
            System.out.println();
            System.out.println("ObjectKey: " + experimentFourKey.toString());
            System.out.println();
            Map<Double, List<ExperimentFourResult>> innerMap = resultsMap.get(experimentFourKey);
            printHeader(latex, separator);
            for (Double spawnRate : innerMap.keySet().stream().sorted().collect(Collectors.toList())) {
                List<ExperimentFourResult> results = innerMap.get(spawnRate);
                AvgExperimentFourResult result = getAvgResult(results);
                if (!latex) {
                    printResult(result.getSpawn(), result.getAvgMet(), result.getAvgChrash(), result.getAvgThroughput(), result.getAvgEk(), result.getPercentChrash(), result.size());
                } else
                    printLatexResult(result.getSpawn(), result.getAvgMet(), result.getAvgChrash(), result.getAvgThroughput(), result.getAvgEk(), result.getPercentChrash(), result.size(), separator);
            }
        }
    }

    public void writeLatexGraphs() {
        for (ExperimentFourKey experimentFourKey : resultsMap.keySet()) {
            Map<Double, List<ExperimentFourResult>> innerMap = resultsMap.get(experimentFourKey);
            for (Double spawnRate : innerMap.keySet()) {

            }
        }
    }

    private AvgExperimentFourResult getAvgResult(List<ExperimentFourResult> results) {
        double spawnRate = results.get(0).getProperties().getSpawnRateNorth();
        AvgExperimentFourResult avgExperimentFourResult = new AvgExperimentFourResult(spawnRate);
        for (ExperimentFourResult result : results) {
            SimulationProperties props = result.getProperties();
            AIMStats stats = result.getStats();
            avgExperimentFourResult.addMET(stats.getMeanEvacTime());
            avgExperimentFourResult.addChrashes(stats.getNofChrashedVehicles());
            avgExperimentFourResult.addThroughput(stats.getThroughput());
            avgExperimentFourResult.addEk(stats.getTotalSpentEk());
            assert props.getSpawnRateNorth() == spawnRate;
        }
        return avgExperimentFourResult;
    }


    private void printHeader(boolean latex, String separator) {
        if (latex)
            if (separator.equals("&")) {
                System.out.println("\\textbf{Spawn Rate} " + separator + " \\textbf{# Runs} " + separator + " \\textbf{MET} " + separator + " \\textbf{Collisions} " + separator + " \\textbf{\\% Collisions} " + separator + " \\textbf{Throughput} " + separator + " \\textbf{Total Kinetic Energy Lost} \\\\ \\hline");
            } else {
                System.out.println("spawn rate" + separator + "nof runs" + separator + "met" + separator + "collisions" + separator + "percent collisions" + separator + "throughput" + separator + "total kinetic energy lost");
            }
        else {
            System.out.println("Spawn Rate    # Runs     MET        Collisions   Percent Collisions   Throughput   Total Kinetic Energy Lost");
        }
    }

    private void printResult(double spawnRate, double meanEvacTime, double nofChrashedVehicles, double throughput, double totalSpentEk, double percentChrash, int nofRuns) {
        System.out.format("%10.6f   ", spawnRate);
        System.out.format("% 10d  ", nofRuns);
        System.out.format("%10.5f   ", meanEvacTime);
        System.out.format("%10.2f   ", nofChrashedVehicles);
        System.out.format("%17.2f  ", percentChrash);
        System.out.format("%10.2f   ", throughput);
        System.out.format("%10.3f   ", totalSpentEk);
        System.out.println();
    }

    private void printLatexResult(double spawnRate, double meanEvacTime, double nofChrashedVehicles, double throughput, double totalSpentEk, double percentChrash, int nofRuns, String separator) {
        System.out.format("% 18.1f  ", spawnRate);
        System.out.format(separator + " %14d  ", nofRuns);
        System.out.format(separator + " %11.2f  ", round(meanEvacTime, 2));
        System.out.format(separator + " %17.2f   ", nofChrashedVehicles);
        System.out.format(separator + " %21.2f  ", percentChrash);
        System.out.format(separator + " %18.2f  ", throughput);
        System.out.format(separator + " %33.2f  ", round(totalSpentEk, 2));
        if (separator.equals("&"))
            System.out.print("\\\\ \\hline");
        System.out.println();
    }

    public double round(double val, int nofDecimals) {
        double decimalConst = Math.pow(10, nofDecimals);
        return Math.round(val * decimalConst) / decimalConst;
    }

    public enum ExperimentFourKey {
        TINY, DT, EK, STARVE, DEFAULT;

    }

    public class ExperimentFourResult {

        private SimulationProperties properties;

        private AIMStats stats;
        private ExperimentFourKey key;


        public ExperimentFourResult(SimulationProperties properties, AIMStats stats) {
            this.properties = properties;
            this.stats = stats;
            this.key = findKey(properties);
        }

        private ExperimentFourKey findKey(SimulationProperties properties) {
            if (!properties.isEvaluateDistanceTravelled())
                return ExperimentFourKey.DT;
            else if (!properties.isEvaluateEk())
                return ExperimentFourKey.EK;
            else if (!properties.isEvaluateStarvation())
                return ExperimentFourKey.STARVE;
            else if (!properties.isEvaluateTinyThroughput())
                return ExperimentFourKey.TINY;
            else return ExperimentFourKey.DEFAULT;
        }


        public SimulationProperties getProperties() {
            return properties;
        }

        public AIMStats getStats() {
            return stats;
        }

        public ExperimentFourKey getKey() {
            return key;
        }

    }

    public static void main(String[] args) throws IOException, JAXBException {
        File experimentsFolder = new File("experiments");
        List<File> experiments = new ArrayList<>();
        experiments.add(new File(experimentsFolder, "ex4fixed"));
//        experiments.put(new File(experimentsFolder, "ex2v9"));
//        experiments.put(new File(experimentsFolder, "ex2v10"));
        ExperimentFourAnalyzer analyzer = new ExperimentFourAnalyzer(experiments);
        analyzer.printResults(true, ",");
        analyzer.printRunningStatistics();
    }
}



