package jossihawk.utils;

import jossihawk.simulator.AIMStats;
import org.jfree.data.xy.XYDataItem;
import org.jfree.data.xy.XYSeries;

import javax.xml.bind.JAXBException;
import java.io.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Properties;

public class ExperimentThreeAnalyzer {

    private List<AIMStats> results;
    private List<Properties> properties;

    public ExperimentThreeAnalyzer(List<File> experiments) throws IOException, JAXBException {
        results = new ArrayList<>();
        properties = new ArrayList<>();
        for (File experimentDir : experiments) {
            File[] resultFiles = experimentDir.listFiles(pathname -> {
                return new File(pathname + File.separator + "results" + File.separator + "globalStats.xml").exists();
            });

            for (int i = 0; i < resultFiles.length; i++) {
                File resultfile = new File(resultFiles[i].getAbsolutePath() + File.separator + "results" + File.separator + "globalStats.xml");
                results.add(AIMStats.readFromFile(resultfile));
                properties.add(readPropertiesFile(new File(resultFiles[i].getAbsolutePath() + File.separator + resultFiles[i].getName() + ".properties")));
            }
        }
    }

    private Properties readPropertiesFile(File propertiesFile) throws IOException {
        FileInputStream fileInput = new FileInputStream(propertiesFile);
        Properties properties = new Properties();
        properties.load(fileInput);
        fileInput.close();
        return properties;
    }

    public void showGraphWithNofCars() {
        List<XYSeries> xySeries = new ArrayList<>();
        xySeries.add(createNofCarsEvaluationsDataset());
        new ChartDrawer(xySeries, "scalingTest");

    }

    public XYSeries createNofCarsEvaluationsDataset() {
        HashMap<Integer, List<Integer>> map = new HashMap<>();
        int tot_datapoints = 0;
        for (AIMStats result : results) {
            if (result.getScalingDataPoints() != null) {
                for (ScalingDataPoint scalingDataPoint : result.getScalingDataPoints()) {
                    tot_datapoints++;
                    int x = scalingDataPoint.getNofCarsInIntersection();
                    int y = scalingDataPoint.getNofEvaluationsBeforeNoCollisions();
                    if (map.containsKey(x)) {
                        map.get(x).add(y);
                    } else {
                        List<Integer> tempList = new ArrayList<>();
                        tempList.add(y);
                        map.put(x, tempList);
                    }
                }
            }
        }
        System.out.println("tot: " + tot_datapoints);
        XYSeries xySeries = new XYSeries("Scaling");
        for (Integer integer : map.keySet()) {
            if(map.get(integer).size()> 10) {
                double averageEvals = avg(map.get(integer));
                xySeries.add((int) integer, averageEvals);
            }
        }
        System.out.println(xySeries.getItemCount());
        return xySeries;
    }

    private double avg(List<Integer> integers) {
        int sum = 0;
        for (Integer integer : integers) {
            sum += integer;
        }
        return sum / ((double) integers.size());
    }

    private XYSeries createSpawnRatesDataset() {
        HashMap<Double, List<Integer>> map = new HashMap<>();
        for (int i = 0; i < results.size(); i++) {
            AIMStats result = results.get(i);
            Properties prop = properties.get(i);
            if (result.getScalingDataPoints() != null) {
                for (ScalingDataPoint scalingDataPoint : result.getScalingDataPoints()) {
                    double x = convertSpawnRate(Double.parseDouble(prop.get("spawnRateNorth").toString()));
                    int y = scalingDataPoint.getNofEvaluationsBeforeNoCollisions();
                    if (map.containsKey(x)) {
                        map.get(x).add(y);
                    } else {
                        List<Integer> tempList = new ArrayList<>();
                        tempList.add(y);
                        map.put(x, tempList);
                    }
                }
            }
        }
        XYSeries xySeries = new XYSeries("Scaling");
        for (Double x : map.keySet()) {
            if (map.get(x).size()> 10) {
                double averageEvals = avg(map.get(x));
                xySeries.add((double) x, averageEvals);
            }
        }
        System.out.println(xySeries.getItemCount());
        return xySeries;
    }

    public double convertSpawnRate(double spawnRate) {
        double[] spawnRates = {
                0.1053383170998845,
                0.22304399484793236,
                0.3564206308646345,
                0.5103040934054393,
                0.6921871619826325,
                0.9146136040714126,
                1.2010783516194223,
                1.6042684308743838,
                2.2920137768080995,
                250.0};
        double returnRate = 0.0;
        for (int i = 0; i < spawnRates.length; i++) {
            double rate = spawnRates[i];
            if(spawnRate == rate)
                returnRate = (i+1)/10.0;
        }
        return returnRate;
    }

    private void showGraphWithSpawnRates() {
        List<XYSeries> xySeries = new ArrayList<>();
        xySeries.add(createSpawnRatesDataset());
        new ChartDrawer(xySeries, "scalingTest");
    }

    public void writeGraph(File graph) throws FileNotFoundException, UnsupportedEncodingException {
        PrintWriter writer = new PrintWriter(graph, "UTF-8");
        XYSeries series = createNofCarsEvaluationsDataset();
        writer.println("steps,evaluations");
        for (int i = 0; i < series.getItemCount(); i++) {
            XYDataItem dataItem = series.getDataItem(i);
            writer.println(dataItem.getXValue() + "," + dataItem.getYValue());
        }
        writer.close();
    }

    public static void main(String[] args) throws IOException, JAXBException {
        List<File> experiments = new ArrayList<>();
        File experimentfolder = new File("experiments");
        experiments.add(new File(experimentfolder, "ex2fixed"));
//        experiments.put(new File(experimentfolder, "ex3v8"));
        ExperimentThreeAnalyzer analyzer = new ExperimentThreeAnalyzer(experiments);

        analyzer.showGraphWithNofCars();
        analyzer.showGraphWithSpawnRates();

        File analyticsFile = new File(experiments.get(0) +File.separator +"analytics");
        if (!analyticsFile.exists())
            analyticsFile.mkdir();

        String graphName = "nofCarsScaling.csv";
        analyzer.writeGraph(new File(analyticsFile, graphName));
    }
}
